import urwid
from util import *


class BigTable(urwid.BoxWidget):
    """
    Table around which the user can navigate. Shows as many columns and rows
    as will fit in the area, scrolling through any surplus as the user moves
    around the selected cell.

    TODO: how this ought to work - pass in widgets rather than text; render
    (and, perforce, keypress via shownGeometry) work out what will fit by
    using rows() and cols() of the canvases rendered by the widgets. render
    then directly puts these canvases together to make its canvas using
    CanvasJoin() and CanvasCombine()
    """
    def __init__(self, legend, colHeaders, rowHeaders, contents,
	    startRow=0, startCol=0, selectedRow=None, selectedCol=None,
	    softGridChar='.', attrOfCell=(lambda i,j: None) ):
	nonempty = len(colHeaders) > 0 and len(rowHeaders) > 0
	if selectedCol == None:
	    if nonempty: selectedCol = 0
	    else: selectedCol = -1
	if selectedRow == None:
	    if nonempty: selectedRow = 0
	    else: selectedRow = -1
	if selectedCol >= len(colHeaders):
	    selectedCol = len(colHeaders) - 1
	if selectedRow >= len(rowHeaders):
	    selectedRow = len(rowHeaders) - 1

	(self.legend, self.colHeaders, self.rowHeaders, self.contents,
		self.startRow, self.startCol, self.selectedRow,
		self.selectedCol, self.softGridChar, self.attrOfCell) = (
			legend, colHeaders, rowHeaders, contents,
			startRow, startCol, selectedRow, selectedCol,
			softGridChar, attrOfCell)

	self.numRows = len(rowHeaders)
	self.numCols = len(colHeaders)

	self.softGridChar = '.'

	# to support old versions of urwid:
	if not hasattr(self, "_invalidate"):
	    setattr(self, "_invalidate", lambda:None)

    def selectable(self):
	return True
    def keypress(self, size, k):
	((startRow, endRow),
		(startCol, endCol),
		rowHeights, colHeadersHeight,
		colWidths, rowHeadersWidth) = self.shownGeometry(size)
	if k in ["down", "ctrl n"]:
	    if self.selectedRow == -1:
		if startRow < endRow:
		    self.selectedRow = startRow
	    else:
		if self.selectedRow < self.numRows-1:
		    self.selectedRow += 1
		    if self.selectedRow >= endRow:
			self.startRow += 1
	    self._invalidate()
	elif k in ["right", "ctrl f"]:
	    if self.selectedCol == -1:
		if self.startCol < endCol:
		    self.selectedCol = self.startCol
	    else:
		if self.selectedCol < self.numCols-1:
		    self.selectedCol += 1
		    if self.selectedCol >= endCol:
			self.startCol += 1
	    self._invalidate()
	elif k in ["up", "ctrl p"]:
	    if self.selectedRow >= 0:
		self.selectedRow -= 1
		if self.selectedRow >= 0 and self.selectedRow < self.startRow:
		    self.startRow = self.selectedRow
		self._invalidate()
	elif k in ["left", "ctrl b"]:
	    if self.selectedCol >= 0:
		self.selectedCol -= 1
		if self.selectedCol >= 0 and self.selectedCol < self.startCol:
		    self.startCol = self.selectedCol
		self._invalidate()
	else:
	    return k
	    
    def _fitBelow(self, list, bound):
	sum = 0
	if list is []: return 0
	for i in range(len(list)):
	    sum += list[i]
	    if sum > bound:
		return i
	return len(list)
    def _textHeight(self, str):
	if str == '': return 0
	else: return len(str.splitlines())
    def _textWidth(self, str):
	if str == '': return 0
	else: return max(map(len, str.splitlines()))
    def shownGeometry(self, (maxcol, maxrow)):
	rowHeights = [ max(map(self._textHeight, [ self.rowHeaders[i] ] + self.contents[i]))
		for i in range(self.startRow, len(self.rowHeaders)) ]
	colHeadersHeight = max(map(self._textHeight, [self.legend] + self.colHeaders))
	rows = self._fitBelow( [ h+1 for h in rowHeights ], maxrow-(colHeadersHeight+1))

	colWidths = [ 2+max(map(self._textWidth, [ self.colHeaders[j] ] +
		[ contentsRow[j] for contentsRow in self.contents]))
		for j in range(self.startCol, len(self.colHeaders)) ]
	rowHeadersWidth = 2+max(map(self._textWidth, [self.legend] + self.rowHeaders))
	cols = self._fitBelow( [ c+1 for c in colWidths ], maxcol-(rowHeadersWidth+1))

	# XXX: side-effect! 
	# If the selected cell is off the edge of what we can show, bring it
	# back in. This probably ought to go somewhere else.
	if self.selectedRow >= self.startRow+rows:
	    if rows > 0:
		self.selectedRow = self.startRow+rows-1
	    else:
		self.selectedRow = -1
	if self.selectedCol >= self.startCol+cols:
	    if cols > 0:
		self.selectedCol = self.startCol+cols-1
	    else:
		self.selectedCol = -1
	if self.selectedRow >= 0 and self.selectedRow < self.startRow:
	    self.selectedRow = self.startRow
	if self.selectedCol >= 0 and self.selectedCol < self.startCol:
	    self.selectedCol = self.startCol

	return ((self.startRow, self.startRow+rows),
		(self.startCol, self.startCol+cols),
		rowHeights, colHeadersHeight,
		colWidths, rowHeadersWidth)

    def render(self, (maxcol, maxrow), focus=False):

	((startRow, endRow),
		(startCol, endCol),
		rowHeights, colHeadersHeight,
		colWidths, rowHeadersWidth) = self.shownGeometry((maxcol,maxrow))

	def widgetOfCell(row, col):
	    if row == -1:
		if col == -1: text = self.legend
		else: text = self.colHeaders[col]
	    else:
		if col == -1: text = self.rowHeaders[row]
		else: text = self.contents[row][col]
	    return urwid.AttrWrap(
		    urwid.Filler( urwid.Text(text, align='center') ),
		    self.attrOfCell(row, col), focus_attr='selected')

	def columnsOfRow(row, arrowsLeft=False, arrowsRight=False):
	    if arrowsLeft: startHeaderBoundaryChar = '<'
	    else: startHeaderBoundaryChar = '|'
	    if arrowsRight: endHeaderBoundaryChar = '>'
	    else: endHeaderBoundaryChar = ' '
	    return urwid.Columns( [
		    ('fixed', rowHeadersWidth, widgetOfCell(row, -1)),
		    ('fixed', 1, urwid.SolidFill(startHeaderBoundaryChar)) ] + intersperse(
		    [ ('fixed', colWidths[i-startCol], widgetOfCell(row, startCol + i))
			for i in range(startCol, endCol) ],
			('fixed', 1, urwid.SolidFill(self.softGridChar))) +
		    [('fixed', 1, urwid.SolidFill(endHeaderBoundaryChar))] )

	if colHeadersHeight+1 > maxrow or rowHeadersWidth+1 > maxcol:
	    return urwid.TextCanvas([" "*maxcol]*maxrow)

	postBarChars = sum([c+1 for c in colWidths])

	if startRow > 0: startHeaderBoundaryChar = '^'
	else: startHeaderBoundaryChar = '-'
	if endRow < self.numRows: endHeaderBoundaryChar = 'v'
	else: endHeaderBoundaryChar = ' '

	pile = urwid.Pile( [
		('fixed', colHeadersHeight, columnsOfRow(-1,
		    arrowsLeft = startCol > 0,
		    arrowsRight = endCol < self.numCols)),
		('fixed', 1,
		    urwid.Filler(urwid.Text(startHeaderBoundaryChar*rowHeadersWidth
			+ '+' + '-'*sum([c+1 for c in colWidths]))))
	] + intersperse(
	    [ ('fixed', rowHeights[i],
		    columnsOfRow(startRow+i))
		    for i in range(endRow-startRow) ],
	('fixed', 1,
	    urwid.Filler(urwid.Text(self.softGridChar*rowHeadersWidth
		+ '|' + self.softGridChar*postBarChars)))) + 
	[('fixed', 1,
		urwid.Filler(urwid.Text(endHeaderBoundaryChar*rowHeadersWidth)))]
	    )
	if focus:
	    if self.selectedRow == -1: cellRow = 0
	    else: cellRow = 1 + self.selectedRow - startRow
	    if self.selectedCol == -1: cellCol = 0
	    else: cellCol = 1 + self.selectedCol - startCol
	    pile.set_focus(cellRow*2)
	    pile.widget_list[cellRow*2].set_focus(cellCol*2)
	return urwid.Filler(pile).render((maxcol,maxrow), focus)

